# (c) Antonin Chambolle Dec 2023
# Python implementation of D. Hochbaum's algorithm for computing the prox
# of the total variation on a graph. Please cite
# D. S. Hochbaum: An efficient algorithm for image segmentation,
#    Markov random fields and related problems. 
#    J. ACM, 48(4):686--701, 2001.
#    https://doi.org/10.1145/502090.502093
# Chambolle, A., Darbon, J.: On Total Variation Minimization and
#    Surface Evolution Using Parametric Maximum Flows. 
#    Int J Comput Vis 84, 288–307 (2009). 
#    https://doi.org/10.1007/s11263-009-0238-9
# Chambolle, A., Darbon, J.: A parametric maximum flow approach 
#    for discrete total variation regularization. 
#    In: Image Processing and Analysis With Graphs: Theory and Practice.
#    CRC Press. (2012)
# Yuri Boykov and Vladimir Kolmogorov: An Experimental Comparison
#    of Min-Cut/Max-Flow Algorithms for Energy Minimization in 
#    Computer Vision.
#    In IEEE Transactions on Pattern Analysis and Machine Intelligence,
#    September 2004.
# if you use this for research. (Boykov-Kolmogorov's graphcut is used,
# any other maxflow solver could be used provided it is fast/easy to extract
# and label the "source" and "sink" partition.)

import numpy as np
import networkx as nx
from networkx.algorithms.flow import boykov_kolmogorov as gcut

def GraphROF(G,precision=1e-3): # G is a DiGraph with capacities and a "value" at each node // this value is then replaced with the proximal point of the total variation on the graph.
    basenodes = list(G.nodes)
    data = dict(G.nodes(data="value", default=0))
    # an interesting variant would use the data only for nodes where it is
    # provided
    d = np.array([data[n] for n in basenodes])

    #R = G.copy() ## do not copy data, create graph with source and sink
    R = G.__class__()
    R.add_nodes_from(basenodes)
    R.add_weighted_edges_from(G.edges(data='capacity'),weight='capacity')
    source = len(basenodes)
    sink = source+1
    R.add_node(source)
    R.add_node(sink)

    # obsolete, list of nodes is faster
    #Gwork = { n: {'label': 0, 'up': False} for n in R.nodes}
    #Gwork[source]['label']=-1
    #Gwork[sink]['label']=-1
    Gwork = { n: {'up': False} for n in R.nodes }
    
    mean = np.mean(d)
    R.add_weighted_edges_from(list((source, n, np.maximum(data[n]-mean,0)) for n in basenodes),weight='capacity')
    R.add_weighted_edges_from(list((n, sink, np.maximum(mean-data[n],0)) for n in basenodes),weight='capacity')

    labels = [0]
    nodelabels = [basenodes] # initially all nodes are labelled 0 but source and sink
    deadlabels = []
    values = [mean]
    
    change=True
    while (change):
            change = False

            R = gcut(R,source,sink)

            source_tree, target_tree = R.graph["trees"]
            partition = (set(source_tree), set(R) - set(source_tree))
            reachable,nonreachable = partition
            for n in reachable:
                Gwork[n]['up']=True
            for n in nonreachable:
                Gwork[n]['up']=False

            # R is not a true residual graph and has to be turned into one
            # by letting R[u][v]['capacity'] -= R[u][v]['flow'] if positive /
            # or += R[v][u]['flow'] if negative
            ed = list(R.edges)
            for e in ed:
                R.edges[e]['capacity'] = R.edges[e]['capacity']-R.edges[e]['flow']
                if Gwork[e[0]]['up'] ^ Gwork[e[1]]['up']:
                    R.remove_edge(*e)
    
            # careful: gcut creates residual capacities from sink to node or node to source
            ed = list(R.in_edges(source))
            R.remove_edges_from(ed)
            ed = list(R.out_edges(sink))
            R.remove_edges_from(ed)
    
            # create links with 0 capacity from source or to sink if not exists
            ed = list(R.out_edges(source,data='capacity'))
            edm = [(source,n,0) for n in basenodes]
            R.add_weighted_edges_from(edm,weight='capacity')
            R.add_weighted_edges_from(ed,weight='capacity')
            ed = list(R.in_edges(sink,data='capacity'))
            edm = [(n,sink,0) for n in basenodes]
            R.add_weighted_edges_from(edm,weight='capacity') 
            R.add_weighted_edges_from(ed,weight='capacity') 
    
            curlab = set(labels) - set(deadlabels)
            for l in curlab:
                #nodl = [n for n in R.nodes if Gwork[n]['label']==l] # should be a list indexed by node
                nodl = nodelabels[l]
                avp = 0
                avm = 0
                nodp = []
                nodm = []
                for n in nodl:
                    if Gwork[n]['up']:
                        nodp.append(n)
                        avp = avp+R.edges[source,n]['capacity']
                    else:
                        nodm.append(n)
                        avm = avm+R.edges[n,sink]['capacity']

                #print (len(nodl),len(nodp),len(nodm),avp,avm)
                if len(nodp)>0:
                    avp = avp/len(nodp)
                if len(nodm)>0:
                    avm = avm/len(nodm)

                if (len(nodp)==0) or (len(nodm)==0) or ((avp < precision) and (avm < precision)):
                    # on a fini : enlever la zone du graphe
                    ed = list(R.in_edges(nodl))
                    R.remove_edges_from(ed)
                    ed = list(R.out_edges(nodl))
                    R.remove_edges_from(ed)
                    for no in nodl:
                        basenodes.remove(no)
                    deadlabels.append(l)
                else:
                    change = True
                    ## update
                    lp = len(labels)
                    lm = l
                    labels.append(lp) # créer un label
                    values.append(values[l] + avp)
                    #values[lp] = values[l] + avp
                    values[lm] = values[l] - avm
                    # set label of all nodes in source to lp
                    #for n in nodp:
                    #    Gwork[n]['label']=lp
                    nodelabels[l] = nodm
                    nodelabels.append(nodp)
                    # update capacities
                    for n in nodm:
                        val = R.edges[n,sink]['capacity']-avm
                        if val > 0:
                            R.edges[n,sink]['capacity'] = val
                            R.edges[source,n]['capacity'] = 0
                        else:
                            R.edges[n,sink]['capacity'] = 0
                            R.edges[source,n]['capacity'] = -val
                    for n in nodp:
                        val = R.edges[source,n]['capacity']-avp
                        if val >0:
                            R.edges[source,n]['capacity'] = val
                            R.edges[n,sink]['capacity'] = 0
                        else:
                            R.edges[source,n]['capacity'] = 0
                            R.edges[n,sink]['capacity'] = -val

    #for n in G.nodes:
    #    G.nodes[n]['value']=values[Gwork[n]['label']]
    for l in range(len(labels)):
        for n in nodelabels[l]:
            G.nodes[n]['value'] = values[l] 
